package org.nakedobjects.nos.remote.command.marshal;

import java.util.Enumeration;

import org.nakedobjects.noa.adapter.Naked;
import org.nakedobjects.noa.adapter.NakedCollection;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.NakedValue;
import org.nakedobjects.noa.adapter.Oid;
import org.nakedobjects.noa.adapter.ResolveState;
import org.nakedobjects.noa.reflect.NakedObjectField;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.util.Assert;
import org.nakedobjects.nof.core.util.UnknownTypeException;
import org.nakedobjects.nof.reflect.remote.data.CollectionData;
import org.nakedobjects.nof.reflect.remote.data.Data;
import org.nakedobjects.nof.reflect.remote.data.KnownObjects;
import org.nakedobjects.nof.reflect.remote.data.ObjectData;
import org.nakedobjects.nof.reflect.remote.data.ReferenceData;
import org.nakedobjects.nof.reflect.remote.data.ValueData;
import org.nakedobjects.nos.remote.command.DataFactory;


/**
 * Utility class to create Data objects representing a graph of NakedObjects.
 * 
 * As each object is serialised its resolved state is changed to SERIALIZING; any object that is marked as
 * SERIALIZING is skipped.
 */
final class BasicObjectSerializer {
    private DataStructure dataStructure;

    public CollectionData serializeCollection(
            final DataFactory factory,
            final NakedCollection collection,
            final int graphDepth,
            final KnownObjects knownObjects) {
        Oid oid = collection.getOid();
        String collectionType = collection.getSpecification().getFullName();
        String elementType = collection.getElementSpecification().getFullName();
        boolean hasAllElements = collection.getResolveState().isTransient() || collection.getResolveState().isResolved();
        ReferenceData[] elements;

        if (hasAllElements) {
            Enumeration e = collection.elements();
            elements = new ReferenceData[collection.size()];
            int i = 0;
            while (e.hasMoreElements()) {
                NakedObject element = (NakedObject) e.nextElement();
                elements[i++] = serializeObject(factory, element, graphDepth, knownObjects);
            }
        } else {
            elements = new ObjectData[0];
        }

        return factory.createCollectionData(collectionType, elementType, oid, elements, hasAllElements, collection.getVersion());
    }

    // TODO rename to serialise.,.,.
    public final ValueData serializeValue(final DataFactory factory, Naked object) {
        if (object == null) {
            return factory.createValueData(null, null);
        } else {
            return factory.createValueData(object.getSpecification().getFullName(), ((NakedValue) object).asEncodedString());
        }
    }

    public final ReferenceData serializeObject(final DataFactory factory, NakedObject object, int depth, KnownObjects knownObjects) {
        Assert.assertNotNull(object);
        return (ReferenceData) serializeObject2(factory, object, depth, knownObjects);
    }

    private final Data serializeObject2(final DataFactory factory, NakedObject object, int graphDepth, KnownObjects knownObjects) {
        Assert.assertNotNull(object);

        ResolveState resolveState = object.getResolveState();
        boolean isTransient = resolveState.isTransient();

        if (!isTransient && (resolveState.isSerializing() || resolveState.isGhost() || graphDepth <= 0)) {
            Assert.assertNotNull("OID needed for reference", object, object.getOid());
            return factory.createIdentityData(object.getSpecification().getFullName(), object.getOid(), object.getVersion());
        }
        if (isTransient && knownObjects.containsKey(object)) {
            return (ObjectData) knownObjects.get(object);
        }

        boolean withCompleteData = resolveState == ResolveState.TRANSIENT || resolveState == ResolveState.RESOLVED;

        String type = object.getSpecification().getFullName();
        Oid oid = object.getOid();
        ObjectData data = factory.createObjectData(type, oid, withCompleteData, object.getVersion());
        if (isTransient) {
            knownObjects.put(object, data);
        }

        NakedObjectField[] fields = dataStructure.getFields(object.getSpecification());
        Data[] fieldContent = new Data[fields.length];
        NakedObjectsContext.getObjectLoader().start(object, object.getResolveState().serializeFrom());
        for (int i = 0; i < fields.length; i++) {
            if (!fields[i].isPersisted()) {
                continue;
            }
            Naked field = fields[i].get(object);

            if (fields[i].isValue()) {
                fieldContent[i] = serializeValue(factory, field);

            } else if (fields[i].isCollection()) {
                fieldContent[i] = serializeCollection(factory, (NakedCollection) field, graphDepth - 1, knownObjects);

            } else if (fields[i].isObject()) {
                if (field == null) {
                    fieldContent[i] = !withCompleteData ? null : factory.createNullData(fields[i].getSpecification()
                            .getFullName());
                } else {
                    fieldContent[i] = serializeObject2(factory, (NakedObject) field, graphDepth - 1, knownObjects);
                }

            } else {
                throw new UnknownTypeException(fields[i]);
            }
        }
        NakedObjectsContext.getObjectLoader().end(object);
        data.setFieldContent(fieldContent);
        return data;
    }

    public void setDataStructure(DataStructure dataStructure) {
        this.dataStructure = dataStructure;
    }

}
// Copyright (c) Naked Objects Group Ltd.
