package org.nakedobjects.nos.remote.command.socket;

import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;

import org.apache.log4j.Logger;
import org.nakedobjects.noa.NakedObjectRuntimeException;
import org.nakedobjects.noa.security.SessionManager;
import org.nakedobjects.noa.security.ConnectionManager;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.system.ServerListener;
import org.nakedobjects.nof.core.util.Assert;
import org.nakedobjects.nof.core.util.DebugInfo;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nof.reflect.remote.ServerDistribution;
import org.nakedobjects.nof.reflect.remote.data.Distribution;
import org.nakedobjects.nof.reflect.remote.data.DistributionLogger;
import org.nakedobjects.nos.remote.command.ConnectionException;
import org.nakedobjects.nos.remote.command.marshal.Marshaller;


public abstract class AbstractServerListener implements ServerListener, DebugInfo {
    private static final Logger LOG = Logger.getLogger(AbstractServerListener.class);
    private static final int PORT = 9567;
    private static final String SERVER_PORT = "server.port";
    private ConnectionManager connectionManager;
    private ServerSocket socket;
    private ProfilingInputStream profilingInputStream;
    private ProfilingOutputStream profilingOutputStream;
    private WorkerPool workerPool;
    private Marshaller encoder;
    private Boolean shutdown = Boolean.FALSE;
    private SessionManager sessionManager;

    public void setSessionManager(final SessionManager sessionManager) {
        this.sessionManager = sessionManager;
    }
    
    public void setConnectionManager(ConnectionManager connectionManager) {
        this.connectionManager = connectionManager;
    }

    public void setEncoder(Marshaller encoder) {
        this.encoder = encoder;
    }

    public void debugData(DebugString debug) {
        debug.appendln("Listener on", socket.toString());
        debug.appendln("Workers", workerPool);
        workerPool.debug(debug);
    }

    public String debugTitle() {
        return "Server Listener";
    }

    protected boolean isShutdown() {
        synchronized (shutdown) {
            return shutdown.booleanValue();
        }
    }

    private void shutdown() {
        if (socket != null) {
            try {
                socket.close();
                socket = null;
                LOG.info("socket closed");
            } catch (IOException e) {
                LOG.error("Failed to close listening socket", e);
            }
        }
        if (workerPool != null) {
            workerPool.shutdown();
            workerPool = null;
            LOG.info("worker pool stopped");
        }
    }

    public void stop() {
        LOG.info("stopping listener");
        synchronized (shutdown) {
            shutdown = Boolean.TRUE;
        }
    }

    public void listen() {
        Assert.assertNotNull("Connection manager needed", connectionManager);
        Assert.assertNotNull("Session manager needed", sessionManager);
        
        workerPool = new WorkerPool(connectionManager, 5);

        ServerDistribution sdd = new ServerDistribution(sessionManager);
        sdd.setEncoder(encoder);

        Distribution sd = new DistributionLogger(encoder, sdd);
        sd.init();

        try {
            int port = NakedObjectsContext.getConfiguration().getInteger(SERVER_PORT, PORT);
            /*
             * TODO server socket to listen on an address and port
             * 
             * String hostAddress = NakedObjectsContext.getConfiguration().getString(ADDRESS); InetAddress
             * address; try { address = hostAddress == null ? null : InetAddress.getByName(hostAddress); int
             * port = NakedObjectsContext.getConfiguration().getInteger(PORT, DEFAULT_PORT); ServerSocket
             * serverSocket = new ServerSocket(port, 2, address); serverSocket.setSoTimeout(5000);
             * 
             * 
             */
            socket = new ServerSocket(port);
            LOG.info("nof listener started on " + socket);
        } catch (IOException e) {
            throw new NakedObjectRuntimeException(e);
        }

        try {
            socket.setSoTimeout(1000);
            LOG.info("listening for connection on " + socket);
            while (!isShutdown()) {
                try {
                    Socket clientSocket = socket.accept();
                    LOG.info("connection accepted from " + clientSocket.getInetAddress());
                    InputStream inputStream = clientSocket.getInputStream();
                    OutputStream outputStream = clientSocket.getOutputStream();
                    if (isDebugging()) {
                        inputStream = profilingInputStream = new ProfilingInputStream(inputStream);
                        outputStream = profilingOutputStream = new ProfilingOutputStream(outputStream);
                    }

                    ServerConnection connection = createClientConnection(inputStream, outputStream, sd);
                    // spawnConnectionThread(connection, sd);

                    Worker worker = workerPool.getWorker();
                    worker.setConnection(connection);
                    // worker.start();

                    // main thread accepts new connection - have a connection - therefore implicitly a request
                    // get worker; associate with connection

                    // receive request
                    // process request
                    // return worker to pool and drop connection
                    // or
                    // repeat from 3 (if KEEP_ALIVE)

                    // main thread accepts new connection
                    // adds connection to to queue
                    // repeats, waiting for new connection

                    // (one of many) consumer picks up connection from queue
                    // wait for request
                    // process request
                    // consumer returns connection
                    // repeat from top

                } catch (InterruptedIOException ignore) {
                    // timeout - check at top of loop 
                    continue; 
                } catch (IOException e) {
                    LOG.warn("connection exception", e);
                    continue; 
                } catch (ConnectionException e) {
                    LOG.warn("connection exception", e);
                }
            }
        } catch (SocketException e) {
            LOG.error("listener failure", e);
        }
        shutdown();
    }

    private boolean isDebugging() {
        return true;
    }

    protected abstract ServerConnection createClientConnection(InputStream input, OutputStream output, Distribution distribution);

}

// Copyright (c) Naked Objects Group Ltd.
